use std::process::{Command, Output};

use regex::Regex;

pub fn update_package_index() -> std::io::Result<Output> {
    Command::new("apt-get").arg("-y").arg("update").output()
}

pub fn run_updates(dry_run: bool) -> std::io::Result<Output> {
    let mut command = Command::new("apt-get");

    if dry_run {
        command.arg("-s");
    }

    command.arg("-y").arg("upgrade").output()
}

pub fn install_packages(packages: &[&str]) -> std::io::Result<Output> {
    Command::new("apt-get")
        .arg("install")
        .arg("-y")
        .args(packages)
        .output()
}

pub fn get_available_updates() -> std::io::Result<Vec<String>> {
    let output = Command::new("apt-get").arg("-sV").arg("upgrade").output()?;
    let upgradables = parse_upgrade_output(&String::from_utf8_lossy(&output.stdout));
    Ok(upgradables)
}

fn parse_upgrade_output(output: &str) -> Vec<String> {
    output
        .split_once("The following packages will be upgraded:\n")
        .and_then(|(_, rest)| {
            // Find the first line with non-whitespace characters (indicating the end of the list)
            let re = Regex::new(r"(?m)^\S").unwrap();
            re.find(rest).map(|m| rest.split_at(m.start()).0)
        })
        .map_or_else(Vec::new, |text| {
            let lines = text.lines();
            lines.map(|line| line.trim().to_owned()).collect()
        })
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn parses_upgrade_output_correctly() {
        // `apt-get -sV upgrade`
        let test_output = r"
NOTE: This is only a simulation!
      apt-get needs root privileges for real execution.
      Keep also in mind that locking is deactivated,
      so don't depend on the relevance to the real current situation!
Reading package lists... Done
Building dependency tree... Done
Reading state information... Done
Calculating upgrade... Done
The following packages have been kept back:
   linux-image-amd64 (5.10.191-1 => 5.10.197-1)
The following packages will be upgraded:
   adduser (3.118 => 3.118+deb11u1)
   base-files (11.1+deb11u7 => 11.1+deb11u8)
   cpio (2.13+dfsg-4 => 2.13+dfsg-7.1~deb11u1)
   dbus (1.12.24-0+deb11u1 => 1.12.28-0+deb11u1)
   distro-info-data (0.51+deb11u3 => 0.51+deb11u4)
   dpkg (1.20.12 => 1.20.13)
   grub-common (2.06-3~deb11u5 => 2.06-3~deb11u6)
   grub-pc (2.06-3~deb11u5 => 2.06-3~deb11u6)
   grub-pc-bin (2.06-3~deb11u5 => 2.06-3~deb11u6)
   grub2-common (2.06-3~deb11u5 => 2.06-3~deb11u6)
   krb5-locales (1.18.3-6+deb11u3 => 1.18.3-6+deb11u4)
   libbsd0 (0.11.3-1 => 0.11.3-1+deb11u1)
   libcurl3-gnutls (7.74.0-1.3+deb11u7 => 7.74.0-1.3+deb11u10)
   libdbus-1-3 (1.12.24-0+deb11u1 => 1.12.28-0+deb11u1)
   libgssapi-krb5-2 (1.18.3-6+deb11u3 => 1.18.3-6+deb11u4)
   libk5crypto3 (1.18.3-6+deb11u3 => 1.18.3-6+deb11u4)
   libkrb5-3 (1.18.3-6+deb11u3 => 1.18.3-6+deb11u4)
   libkrb5support0 (1.18.3-6+deb11u3 => 1.18.3-6+deb11u4)
   libncurses6 (6.2+20201114-2+deb11u1 => 6.2+20201114-2+deb11u2)
   libncursesw6 (6.2+20201114-2+deb11u1 => 6.2+20201114-2+deb11u2)
   libnss-systemd (247.3-7+deb11u2 => 247.3-7+deb11u4)
   libpam-systemd (247.3-7+deb11u2 => 247.3-7+deb11u4)
   libssl1.1 (1.1.1n-0+deb11u5 => 1.1.1w-0+deb11u1)
   libsystemd0 (247.3-7+deb11u2 => 247.3-7+deb11u4)
   libtinfo6 (6.2+20201114-2+deb11u1 => 6.2+20201114-2+deb11u2)
   libudev1 (247.3-7+deb11u2 => 247.3-7+deb11u4)
   logrotate (3.18.0-2+deb11u1 => 3.18.0-2+deb11u2)
   ncurses-base (6.2+20201114-2+deb11u1 => 6.2+20201114-2+deb11u2)
   ncurses-bin (6.2+20201114-2+deb11u1 => 6.2+20201114-2+deb11u2)
   ncurses-term (6.2+20201114-2+deb11u1 => 6.2+20201114-2+deb11u2)
   openssh-client (1:8.4p1-5+deb11u1 => 1:8.4p1-5+deb11u2)
   openssh-server (1:8.4p1-5+deb11u1 => 1:8.4p1-5+deb11u2)
   openssh-sftp-server (1:8.4p1-5+deb11u1 => 1:8.4p1-5+deb11u2)
   openssl (1.1.1n-0+deb11u5 => 1.1.1w-0+deb11u1)
   qemu-utils (1:5.2+dfsg-11+deb11u2 => 1:5.2+dfsg-11+deb11u3)
   systemd (247.3-7+deb11u2 => 247.3-7+deb11u4)
   systemd-sysv (247.3-7+deb11u2 => 247.3-7+deb11u4)
   udev (247.3-7+deb11u2 => 247.3-7+deb11u4)
38 upgraded, 0 newly installed, 0 to remove and 1 not upgraded.
Inst base-files [11.1+deb11u7] (11.1+deb11u8 Debian:11.8/oldstable [amd64])
Conf base-files (11.1+deb11u8 Debian:11.8/oldstable [amd64])
Inst dpkg [1.20.12] (1.20.13 Debian:11.8/oldstable [amd64])
Conf dpkg (1.20.13 Debian:11.8/oldstable [amd64])
Inst ncurses-bin [6.2+20201114-2+deb11u1] (6.2+20201114-2+deb11u2 Debian:11.8/oldstable [amd64])
Conf ncurses-bin (6.2+20201114-2+deb11u2 Debian:11.8/oldstable [amd64])
Inst ncurses-base [6.2+20201114-2+deb11u1] (6.2+20201114-2+deb11u2 Debian:11.8/oldstable [all])
Conf ncurses-base (6.2+20201114-2+deb11u2 Debian:11.8/oldstable [all])
Inst libnss-systemd [247.3-7+deb11u2] (247.3-7+deb11u4 Debian:11.8/oldstable, Debian:11-updates/oldstable-updates [amd64]) []
Inst libsystemd0 [247.3-7+deb11u2] (247.3-7+deb11u4 Debian:11.8/oldstable, Debian:11-updates/oldstable-updates [amd64]) [systemd:amd64 ]
Conf libsystemd0 (247.3-7+deb11u4 Debian:11.8/oldstable, Debian:11-updates/oldstable-updates [amd64]) [systemd:amd64 ]
Inst libpam-systemd [247.3-7+deb11u2] (247.3-7+deb11u4 Debian:11.8/oldstable, Debian:11-updates/oldstable-updates [amd64]) [systemd:amd64 ]
Inst systemd [247.3-7+deb11u2] (247.3-7+deb11u4 Debian:11.8/oldstable, Debian:11-updates/oldstable-updates [amd64])
Inst udev [247.3-7+deb11u2] (247.3-7+deb11u4 Debian:11.8/oldstable, Debian:11-updates/oldstable-updates [amd64]) []
Inst libudev1 [247.3-7+deb11u2] (247.3-7+deb11u4 Debian:11.8/oldstable, Debian:11-updates/oldstable-updates [amd64])
Conf libudev1 (247.3-7+deb11u4 Debian:11.8/oldstable, Debian:11-updates/oldstable-updates [amd64])
Inst adduser [3.118] (3.118+deb11u1 Debian:11.8/oldstable [all])
Conf adduser (3.118+deb11u1 Debian:11.8/oldstable [all])
Conf systemd (247.3-7+deb11u4 Debian:11.8/oldstable, Debian:11-updates/oldstable-updates [amd64])
Inst systemd-sysv [247.3-7+deb11u2] (247.3-7+deb11u4 Debian:11.8/oldstable, Debian:11-updates/oldstable-updates [amd64])
Inst dbus [1.12.24-0+deb11u1] (1.12.28-0+deb11u1 Debian:11.8/oldstable [amd64]) []
Inst libdbus-1-3 [1.12.24-0+deb11u1] (1.12.28-0+deb11u1 Debian:11.8/oldstable [amd64])
Inst libk5crypto3 [1.18.3-6+deb11u3] (1.18.3-6+deb11u4 Debian:11.8/oldstable [amd64])
Conf libk5crypto3 (1.18.3-6+deb11u4 Debian:11.8/oldstable [amd64])
Inst libkrb5support0 [1.18.3-6+deb11u3] (1.18.3-6+deb11u4 Debian:11.8/oldstable [amd64]) [libkrb5-3:amd64 ]
Conf libkrb5support0 (1.18.3-6+deb11u4 Debian:11.8/oldstable [amd64]) [libkrb5-3:amd64 ]
Inst libkrb5-3 [1.18.3-6+deb11u3] (1.18.3-6+deb11u4 Debian:11.8/oldstable [amd64]) [libgssapi-krb5-2:amd64 ]
Conf libkrb5-3 (1.18.3-6+deb11u4 Debian:11.8/oldstable [amd64]) [libgssapi-krb5-2:amd64 ]
Inst libgssapi-krb5-2 [1.18.3-6+deb11u3] (1.18.3-6+deb11u4 Debian:11.8/oldstable [amd64])
Conf libgssapi-krb5-2 (1.18.3-6+deb11u4 Debian:11.8/oldstable [amd64])
Inst libssl1.1 [1.1.1n-0+deb11u5] (1.1.1w-0+deb11u1 Debian:11.8/oldstable [amd64])
Conf libssl1.1 (1.1.1w-0+deb11u1 Debian:11.8/oldstable [amd64])
Inst libncurses6 [6.2+20201114-2+deb11u1] (6.2+20201114-2+deb11u2 Debian:11.8/oldstable [amd64]) []
Inst libncursesw6 [6.2+20201114-2+deb11u1] (6.2+20201114-2+deb11u2 Debian:11.8/oldstable [amd64]) []
Inst libtinfo6 [6.2+20201114-2+deb11u1] (6.2+20201114-2+deb11u2 Debian:11.8/oldstable [amd64])
Conf libtinfo6 (6.2+20201114-2+deb11u2 Debian:11.8/oldstable [amd64])
Inst cpio [2.13+dfsg-4] (2.13+dfsg-7.1~deb11u1 Debian:11.8/oldstable [amd64])
Inst logrotate [3.18.0-2+deb11u1] (3.18.0-2+deb11u2 Debian:11.8/oldstable [amd64])
Inst krb5-locales [1.18.3-6+deb11u3] (1.18.3-6+deb11u4 Debian:11.8/oldstable [all])
Inst ncurses-term [6.2+20201114-2+deb11u1] (6.2+20201114-2+deb11u2 Debian:11.8/oldstable [all])
Inst openssh-sftp-server [1:8.4p1-5+deb11u1] (1:8.4p1-5+deb11u2 Debian:11.8/oldstable [amd64]) []
Inst openssh-server [1:8.4p1-5+deb11u1] (1:8.4p1-5+deb11u2 Debian:11.8/oldstable [amd64]) []
Inst openssh-client [1:8.4p1-5+deb11u1] (1:8.4p1-5+deb11u2 Debian:11.8/oldstable [amd64])
Inst distro-info-data [0.51+deb11u3] (0.51+deb11u4 Debian:11.8/oldstable [all])
Inst grub2-common [2.06-3~deb11u5] (2.06-3~deb11u6 Debian-Security:11/oldstable-security [amd64]) [grub-pc:amd64 ]
Inst grub-pc [2.06-3~deb11u5] (2.06-3~deb11u6 Debian-Security:11/oldstable-security [amd64]) []
Inst grub-pc-bin [2.06-3~deb11u5] (2.06-3~deb11u6 Debian-Security:11/oldstable-security [amd64]) []
Inst grub-common [2.06-3~deb11u5] (2.06-3~deb11u6 Debian-Security:11/oldstable-security [amd64])
Inst libbsd0 [0.11.3-1] (0.11.3-1+deb11u1 Debian:11.8/oldstable [amd64])
Inst libcurl3-gnutls [7.74.0-1.3+deb11u7] (7.74.0-1.3+deb11u10 Debian-Security:11/oldstable-security [amd64])
Inst openssl [1.1.1n-0+deb11u5] (1.1.1w-0+deb11u1 Debian:11.8/oldstable [amd64])
Inst qemu-utils [1:5.2+dfsg-11+deb11u2] (1:5.2+dfsg-11+deb11u3 Debian:11.8/oldstable [amd64])
Conf libnss-systemd (247.3-7+deb11u4 Debian:11.8/oldstable, Debian:11-updates/oldstable-updates [amd64])
Conf libpam-systemd (247.3-7+deb11u4 Debian:11.8/oldstable, Debian:11-updates/oldstable-updates [amd64])
Conf udev (247.3-7+deb11u4 Debian:11.8/oldstable, Debian:11-updates/oldstable-updates [amd64])
Conf systemd-sysv (247.3-7+deb11u4 Debian:11.8/oldstable, Debian:11-updates/oldstable-updates [amd64])
Conf dbus (1.12.28-0+deb11u1 Debian:11.8/oldstable [amd64])
Conf libdbus-1-3 (1.12.28-0+deb11u1 Debian:11.8/oldstable [amd64])
Conf libncurses6 (6.2+20201114-2+deb11u2 Debian:11.8/oldstable [amd64])
Conf libncursesw6 (6.2+20201114-2+deb11u2 Debian:11.8/oldstable [amd64])
Conf cpio (2.13+dfsg-7.1~deb11u1 Debian:11.8/oldstable [amd64])
Conf logrotate (3.18.0-2+deb11u2 Debian:11.8/oldstable [amd64])
Conf krb5-locales (1.18.3-6+deb11u4 Debian:11.8/oldstable [all])
Conf ncurses-term (6.2+20201114-2+deb11u2 Debian:11.8/oldstable [all])
Conf openssh-sftp-server (1:8.4p1-5+deb11u2 Debian:11.8/oldstable [amd64])
Conf openssh-server (1:8.4p1-5+deb11u2 Debian:11.8/oldstable [amd64])
Conf openssh-client (1:8.4p1-5+deb11u2 Debian:11.8/oldstable [amd64])
Conf distro-info-data (0.51+deb11u4 Debian:11.8/oldstable [all])
Conf grub2-common (2.06-3~deb11u6 Debian-Security:11/oldstable-security [amd64])
Conf grub-pc (2.06-3~deb11u6 Debian-Security:11/oldstable-security [amd64])
Conf grub-pc-bin (2.06-3~deb11u6 Debian-Security:11/oldstable-security [amd64])
Conf grub-common (2.06-3~deb11u6 Debian-Security:11/oldstable-security [amd64])
Conf libbsd0 (0.11.3-1+deb11u1 Debian:11.8/oldstable [amd64])
Conf libcurl3-gnutls (7.74.0-1.3+deb11u10 Debian-Security:11/oldstable-security [amd64])
Conf openssl (1.1.1w-0+deb11u1 Debian:11.8/oldstable [amd64])
Conf qemu-utils (1:5.2+dfsg-11+deb11u3 Debian:11.8/oldstable [amd64])
            ";

        let upgradables = parse_upgrade_output(test_output);
        assert_eq!(upgradables.len(), 38);
    }
}
