use std::{pin::Pin, process::Stdio, sync::Mutex};

use tokio_stream::{
    wrappers::{ReceiverStream, WatchStream},
    Stream, StreamExt,
};
use tokio_util::codec::{BytesCodec, FramedRead};
use tonic::{Request, Response, Status};

use crate::{debian, health::HealthMonitor, info::Info, task::TaskBuilder};

use super::proto::*;

type AgentResult<T> = std::result::Result<Response<T>, Status>;

pub struct AgentService<'a> {
    pub health: HealthMonitor,
    pub info: &'a Mutex<Info>, // TODO: Find a way to remove the Mutex dependency here
}

#[tonic::async_trait]
impl agent_server::Agent for AgentService<'static> {
    type HealthStream = Pin<Box<dyn Stream<Item = Result<HealthResponse, Status>> + Send>>;

    async fn health(&self, _: Request<()>) -> AgentResult<Self::HealthStream> {
        let receiver = self.health.monitor();
        let version = env!("CARGO_PKG_VERSION");

        let output = WatchStream::new(receiver).map(|health| {
            Ok(HealthResponse {
                version: version.to_owned(),
                system: Some(health.system().into()),
                tasks: health
                    .tasks()
                    .iter()
                    .map(|(k, v)| (k.clone(), v.into()))
                    .collect(),
            })
        });

        Ok(Response::new(Box::pin(output)))
    }

    async fn get_sys_info(&self, _: Request<()>) -> AgentResult<SysInfoResponse> {
        Ok(Response::new(SysInfoResponse::from(
            &*self.info.lock().unwrap(),
        )))
    }

    type SysUpdateStream = Pin<Box<dyn Stream<Item = Result<SysUpdateResponse, Status>> + Send>>;

    async fn sys_update(
        &self,
        req: Request<SysUpdateRequest>,
    ) -> AgentResult<Self::SysUpdateStream> {
        let dry_run = req.get_ref().dry_run;

        let mut receiver =
            TaskBuilder::new("system update".to_owned()).health_monitor(self.health.clone());

        if dry_run {
            receiver = receiver
                .add_step(async { Ok("simulating a system update...".to_owned()) })
                .add_step(async {
                    const DUR: std::time::Duration = std::time::Duration::from_secs(5);
                    tokio::time::sleep(DUR).await;
                    Ok("completed running an artifical delay...".to_owned())
                });
        }

        let receiver = receiver
            .add_step(async move {
                tokio::task::spawn_blocking(move || {
                    let output = debian::run_updates(dry_run).map_err(|err| {
                        tracing::error!(%err, "failed to run updates");
                        err
                    })?;

                    let out = if !output.status.success() {
                        tracing::error!(?output, "child process exited unsuccessfuly");

                        match output.status.code() {
                            Some(exit_code) => Err(Status::internal(format!(
                                "operation exited with error (code {exit_code})"
                            ))),
                            None => Err(Status::cancelled("operation was cancelled by signal")),
                        }
                    } else {
                        Ok(String::from_utf8_lossy(output.stdout.as_slice()).to_string())
                    };

                    // TODO: We could split the output by lines and emit those as "steps" so the
                    // upgrade process is more interactive
                    out
                })
                .await
                .unwrap()
            })
            .build()
            .into_background();

        let stream = ReceiverStream::new(receiver).map(|output| {
            output
                .map(|output| SysUpdateResponse {
                    output,
                    progress: 1,
                })
                .map_err(|err| Status::internal(err.to_string()))
        });

        Ok(Response::new(Box::pin(stream)))
    }

    type ExecStream = Pin<Box<dyn Stream<Item = Result<ExecResponse, Status>> + Send>>;

    async fn exec(&self, req: Request<ExecRequest>) -> AgentResult<Self::ExecStream> {
        use exec_response::Out;

        let ExecRequest { program, args } = req.get_ref();

        let mut command = tokio::process::Command::new(program)
            .args(args)
            .stdout(Stdio::piped())
            .stderr(Stdio::piped())
            .spawn()?;

        let stdout =
            FramedRead::new(command.stdout.take().unwrap(), BytesCodec::new()).map(|stdout| {
                let stdout = stdout.unwrap();
                Out::Stdout(String::from_utf8_lossy(&stdout[..]).to_string())
            });

        let stderr =
            FramedRead::new(command.stderr.take().unwrap(), BytesCodec::new()).map(|stderr| {
                let stderr = stderr.unwrap();
                Out::Stderr(String::from_utf8_lossy(&stderr[..]).to_string())
            });

        let exit = TaskBuilder::new(format!("exec {program}"))
            .health_monitor(self.health.clone())
            .add_step(async move { command.wait().await.unwrap() })
            .build()
            .into_stream();

        let stream = stdout
            .merge(stderr)
            .chain(exit.map(|code| Out::ExitCode(code.code().unwrap_or_default())))
            .map(|out| Ok(ExecResponse { out: Some(out) }));

        Ok(Response::new(Box::pin(stream)))
    }
}
