use std::time::Duration;

use tokio::{signal, sync::oneshot};
use tower_http::trace::TraceLayer;

use crate::{
    health::HealthMonitor,
    info,
    server::{agent::AgentService, proto::agent_server},
};

mod agent;
mod proto {
    tonic::include_proto!("prymn");

    impl From<&crate::health::SystemHealth> for SystemHealth {
        fn from(val: &crate::health::SystemHealth) -> Self {
            if let crate::health::SystemStatus::Critical(ref reasons) = val.status {
                SystemHealth {
                    status: itertools::join(reasons.iter().map(ToString::to_string), ","),
                }
            } else {
                SystemHealth {
                    status: val.status.to_string(),
                }
            }
        }
    }

    impl From<&crate::task::TaskStatus> for TaskHealth {
        fn from(value: &crate::task::TaskStatus) -> Self {
            Self {
                started_on: value.started_on().to_string(),
                progress: value.progress(),
            }
        }
    }

    impl From<&crate::info::Info> for SysInfoResponse {
        fn from(info: &crate::info::Info) -> Self {
            use sysinfo::{CpuExt, DiskExt, SystemExt};

            let system = info.system();

            let cpus = system
                .cpus()
                .iter()
                .map(|cpu| sys_info_response::Cpu {
                    freq_mhz: cpu.frequency(),
                    usage: cpu.cpu_usage(),
                })
                .collect();

            let disks = system
                .disks()
                .iter()
                .map(|disk| sys_info_response::Disk {
                    name: disk.name().to_string_lossy().into_owned(),
                    total_bytes: disk.total_space(),
                    avail_bytes: disk.available_space(),
                    mount_point: disk.mount_point().to_string_lossy().into_owned(),
                })
                .collect();

            Self {
                uptime: system.uptime(),
                hostname: system.host_name().unwrap_or_default(),
                os: system.long_os_version().unwrap_or_default(),
                mem_total_bytes: system.total_memory(),
                mem_avail_bytes: system.available_memory(),
                swap_total_bytes: system.total_swap(),
                swap_free_bytes: system.free_swap(),
                updates_available: info.updates().len() as u32,
                cpus,
                disks,
            }
        }
    }
}

/// Run the server. This is the main entry point of the application.
#[tokio::main]
pub async fn run() -> anyhow::Result<()> {
    let (shutdown_tx, shutdown_rx) = oneshot::channel();

    // Listen for shutdown signals
    tokio::spawn(async {
        signal::ctrl_c()
            .await
            .expect("failed to listen to a ctrl-c signal");

        let _ = shutdown_tx.send(());
    });

    let info = info::spawn_info_subsystem();
    let health_monitor = HealthMonitor::new();

    // Monitor system info forever
    // TODO: Maybe we can move it inside the server response function?
    // We could spawn a new loop whenever we need it, but the problem is when does it get
    // destroyed?
    {
        let health_monitor = health_monitor.clone();
        tokio::spawn(async move {
            loop {
                health_monitor.check_system_info(&info.lock().unwrap());
                tokio::time::sleep(Duration::from_secs(1)).await;
            }
        });
    }

    let agent_service = agent_server::AgentServer::new(AgentService {
        health: health_monitor.clone(),
        info,
    });

    let addr = "[::]:50012".parse()?;
    tracing::info!("listening on {}", addr);
    tonic::transport::Server::builder()
        .layer(TraceLayer::new_for_grpc())
        .add_service(agent_service)
        .serve_with_shutdown(addr, async {
            let _ = shutdown_rx.await;
        })
        .await?;

    Ok(())
}
