//! A task is an atomic executing routine that the agent is running, potentially in the background.
//! The task is tracked by the system monitor.

// TODO: Take a look at futures::stream::FuturesOrdered
//       It is used to store futures in an ordered fashion, and it also implements Stream

use std::{
    future::Future,
    pin::Pin,
    task::{Context, Poll},
};

use chrono::{DateTime, Utc};
use tokio::sync::{mpsc, watch};
use tokio_stream::{Stream, StreamExt};

use super::health::HealthMonitor;

#[derive(Clone, Default)]
pub struct TaskStatus {
    started_on: DateTime<Utc>,
    curr_step: usize,
    max_steps: usize,
}

impl TaskStatus {
    /// Returns the task progress as a percentage value
    pub fn progress(&self) -> f32 {
        100.0 * (self.curr_step as f32 / self.max_steps as f32)
    }

    /// Returns the datetime when this task began executing
    pub fn started_on(&self) -> &DateTime<Utc> {
        &self.started_on
    }

    fn next_step(&mut self) {
        if self.curr_step < self.max_steps {
            self.curr_step += 1;
        }
    }
}

type BoxFuture<T> = Pin<Box<dyn Future<Output = T> + Send>>;

pub struct TaskBuilder<Step> {
    task: Task<Step>,
}

impl<T> TaskBuilder<T> {
    pub fn new(name: String) -> Self {
        let (sender, _) = watch::channel(TaskStatus::default());

        Self {
            task: Task {
                name,
                health_monitor: None,
                status_channel: sender,
                steps: Vec::new(),
            },
        }
    }

    /// Attaches a health monitor to notify the health system on progress made.
    pub fn health_monitor(mut self, health_monitor: HealthMonitor) -> Self {
        self.task.health_monitor = Some(health_monitor);
        self
    }

    pub fn build(self) -> Task<T> {
        self.task
    }
}

impl<T: Send + 'static> TaskBuilder<BoxFuture<T>> {
    pub fn add_step(mut self, step: impl Future<Output = T> + Send + 'static) -> Self {
        self.task.add_step(step);
        self
    }
}

pub struct Task<T> {
    name: String,
    health_monitor: Option<HealthMonitor>,
    status_channel: watch::Sender<TaskStatus>,
    steps: Vec<T>,
}

impl<T: Send + 'static> Task<BoxFuture<T>> {
    fn add_step(&mut self, step: impl Future<Output = T> + Send + 'static) {
        self.steps.push(Box::pin(step))
    }

    /// Turn this Task into an object that implements [Stream].
    ///
    /// The new stream will output each step's future output.
    pub fn into_stream(self) -> TaskStream<T> {
        if let Some(health) = &self.health_monitor {
            health.track_task(self.name.clone(), self.status_channel.subscribe());
        }

        // Immediately notify the initial status (step 0)
        self.status_channel.send_replace(TaskStatus {
            started_on: Utc::now(),
            curr_step: 0,
            max_steps: self.steps.len(),
        });

        TaskStream { inner: self }
    }

    /// Run this task concurrently in the background.
    ///
    /// Returns a [mpsc::Receiver<T>] which receives the returned values of each step's future
    /// output.
    pub fn into_background(self) -> mpsc::Receiver<T> {
        let (sender, receiver) = mpsc::channel(10);

        tokio::spawn(async move {
            let mut stream = self.into_stream();
            while let Some(value) = stream.next().await {
                let _ = sender.send(value).await;
            }
        });

        receiver
    }
}

pub struct TaskStream<T> {
    inner: Task<BoxFuture<T>>,
}

impl<T> Stream for TaskStream<T> {
    type Item = T;

    fn poll_next(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Option<Self::Item>> {
        match self.inner.steps.get_mut(0) {
            Some(fut) => match fut.as_mut().poll(cx) {
                Poll::Ready(value) => {
                    self.inner.steps.remove(0);

                    self.inner
                        .status_channel
                        .send_modify(|task| task.next_step());

                    Poll::Ready(Some(value))
                }
                Poll::Pending => Poll::Pending,
            },
            None => Poll::Ready(None),
        }
    }
}
